#include "light.h"

/*****************************************************************************/
// Determines if the point is inside the polygon using the 'left-of-line' rule
// ie, if the point is left of every edge in the convex polygon, then it is inside
// it
/*****************************************************************************/

unsigned char pointInPolygon(vector point, polygon poly)
{
	vector v, t, p, zv(0.0, 0.0, 1.0);

	v = poly.vertex2 - poly.vertex1;

	t = v % poly.plane_normal;
	
	p = point - poly.vertex1;
	
	if (dot(p, t) <= 0.0)
	{
		v = poly.vertex3 - poly.vertex2;
		
		t = v % poly.plane_normal;
		
		p = point - poly.vertex2;
		
		if (dot(p, t) <= 0.0)
		{
			v = poly.vertex1 - poly.vertex3;
			
			t = v % poly.plane_normal;
			
			p = point - poly.vertex3;

			if (dot(p, t) <= 0.0)
			{
				return TRUE;
			}
		}
	}

	return FALSE;
}

/*****************************************************************************/
// Determines if a ray cast from ray_origin intersects the polygon.
// It finds the point of intersection with the polygons plane, then does a test
// to determine if the point is inside the polygon (see function above)
/*****************************************************************************/

unsigned char doesRayIntersectPolygon(vector ray, vector ray_origin, polygon poly)
{
	double d = dot(ray, poly.plane_normal);
	if (d < 0.0)
	{
		vector p;
		double t, n;

		p = ray_origin - poly.vertex1;

		n = dot(p, poly.plane_normal);

		if (n > 0.0)
		{
			d = -d;

			t = n / d;

			if (t < 0.9 && t > 0.0)
			{
				vector poi;
				
				poi = ray_origin + (ray * t);
				
				if (pointInPolygon(poi, poly) == TRUE)
				{
					return TRUE;
				}
			}
		}
	}

	return FALSE;
}

/*****************************************************************************/
// Traces a a ray from ray_origin and compares it to every polygon in the scene
// until it detects an intersection
/*****************************************************************************/

unsigned char rayTrace(vector ray, vector ray_origin, object **object_list, int object_count)
{
	unsigned char tof = FALSE;

	int i = 0, j = 0;

	while (i < object_count && tof == FALSE)
	{
		j = 0;

		while (j < object_list[i]->polygon_count && tof == FALSE)
		{
			tof = doesRayIntersectPolygon(ray, ray_origin, object_list[i]->polygons[j]);
			j++;
		}
		i++;
	}

	return tof;
}

/*****************************************************************************/
// getLightDirection() returns the direction vector of the light to a point for
// each type of light source
/*****************************************************************************/

vector light::getLightDirection(vector nothing)
{
	// NULL function
}

vector pointlight::getLightDirection(vector pos)
{
	vector dir;

	dir = position - pos;

	return dir;
}

vector spotlight::getLightDirection(vector pos)
{
	vector dir;

	dir = position - pos;

	return dir;
}

vector directedlight::getLightDirection(vector nothing)
{
	return -direction;
}

/*****************************************************************************/
// getLightDirection() returns the intensity of the light reaching a point for
// each type of light source
// Using ray-trace functions above, can be used to determine shadows, but this is
// removed in this case (shadow volumes used instead)
/*****************************************************************************/

colour light::getLightIntensity(vector nothing)
{
	// NULL function
}

colour pointlight::getLightIntensity(vector pos)
{
	colour intens;

	if (linear_attenuation != -1)
	{
		double distance;
		vector d;

		d = position - pos;

		distance = d.mag();

		if (distance < linear_attenuation)
		{
			double a = distance / linear_attenuation;
			intens = intensity * a;
		} else {
			intens.r = 0.0;
			intens.g = 0.0;
			intens.b = 0.0;
		}
	} else {
		intens = intensity;
	}

	return intens;
}

colour spotlight::getLightIntensity(vector pos)
{
	colour intens;
	double distance, angle_between;
	vector d, c, r;

	d = pos - position;

	distance = d.mag();

	if (distance < max_distance)
	{
		vector up(0.0, 1.0, 0.0);
		r = d;
		d = d / distance;
		c = direction.normalised();

		angle_between = dot(d, c);

		if (angle_between > penumbra)
		{
			if (angle_between < umbra)
			{
				double b = (angle_between - penumbra) / (umbra - penumbra);
				intens = intensity * b;
				return intens;
			} else {
				return intensity;
			}
		}
	}
	
	return intens;
}

colour directedlight::getLightIntensity(vector nothing)
{
	return intensity;
}

/*****************************************************************************/
// Following functions set the position, direction and intensities of the light sources
/*****************************************************************************/

void light::setPosition(vector nothing)
{
	// NULL function
}

void pointlight::setPosition(vector pos)
{
	position = pos;
}

void spotlight::setPosition(vector pos)
{
	position = pos;
}
	
void light::setDirection(vector nothing)
{
	// NULL function
}

void directedlight::setDirection(vector dir)
{
	direction = dir;
}

void spotlight::setDirection(vector dir)
{
	direction = dir;
}

void light::setAttenuation(double f)
{
	// NULL function
}

void pointlight::setAttenuation(double f)
{
	linear_attenuation = f;
}

void light::setIntensity(colour col)
{
	intensity = col;
}

/*****************************************************************************/
// drawLight() draws the light source in the display
/*****************************************************************************/

void light::drawLight(void)
{
	// NULL function
}

void pointlight::drawLight(void)
{
	double n = 1.0 / 255;

	glDisable(GL_TEXTURE_2D);
	glColor4f(intensity.r * n, intensity.g * n, intensity.b * n, 1.0);
	glPushMatrix();
	glTranslatef(position.x, position.y, position.z);
	glutSolidSphere(0.5, 10, 10);
	glPopMatrix();
	glEnable(GL_TEXTURE_2D);
	glColor4f(1.0, 1.0, 1.0, 1.0);
}

void directedlight::drawLight(void)
{
	double n = 1.0 / 255;

	glDisable(GL_TEXTURE_2D);
	glColor4f(intensity.r * n, intensity.g * n, intensity.b * n, 1.0);
	glBegin(GL_LINES);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(-direction.x * 10, -direction.y * 10, -direction.z * 10);
	glEnd();
	glEnable(GL_TEXTURE_2D);
	glColor4f(1.0, 1.0, 1.0, 1.0);
}

void spotlight::drawLight(void)
{
	double n = 1.0 / 255;

	glDisable(GL_TEXTURE_2D);
	glPushMatrix();
	glRotatef(-90.0, 1.0, 0.0, 0.0);
	glRotatef(-90.0, 0.0, 1.0, 0.0);
	gluLookAt(position.x, position.y, position.z, direction.x, direction.y, direction.z, 0.0, 1.0, 0.0);
	glColor4f(0.0, 0.0, 0.0, 1.0);
	glutSolidCone(umbra + 0.1, sin(acos(umbra)), 10, 1);
	glColor4f(intensity.r * n, intensity.g * n, intensity.b * n, 1.0);
	glutSolidCone(umbra, sin(acos(umbra)), 10, 1);
	glPopMatrix();

	glEnable(GL_TEXTURE_2D);
	glColor4f(1.0, 1.0, 1.0, 1.0);
}

/*****************************************************************************/
// generate ShadowMap() was used for tests in creating shadow maps for spot lights
// did not work correctly so was abandoned, but here it is anyway
/*****************************************************************************/

void light::generateShadowMap(object **object_list, int object_count)
{
	// NULL function
}

void spotlight::generateShadowMap(object **object_list, int object_count)
{
	int shadow_map_size = shadow_map_x * shadow_map_y;
	int viewport[4];

// Clear the shadow map

	if (shadow_map != NULL)
	{
		delete [] shadow_map;
	}
	shadow_map = new float [shadow_map_size];

// Determine the angle for the field of vision of the spotlight and clamp to under 180 degrees

	double spot_angle = (acos(penumbra) * (180.0 / 3.141592654) * 2.0);
	if (spot_angle >= 179.0) spot_angle = 179.0;

	glClear(GL_DEPTH_BUFFER_BIT);

// Sets up the projection matrix to render the shadow map from the lights viewpoint

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	gluPerspective(spot_angle, 1.0, min_distance, max_distance);
	glViewport(0, 0, shadow_map_x, shadow_map_y);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glGetDoublev(GL_MODELVIEW_MATRIX, modelview_matrix.m);
	gluLookAt(position.x, position.y, position.z, position.x + direction.x, position.y + direction.y, position.z + direction.z, 0.0, 1.0, 0.0);

	glDisable(GL_TEXTURE_2D);
	glDisable(GL_CULL_FACE);

// Draw all the objects using no lighting/texture information (only interested in depth buffer)

	for (int i = 0; i < object_count; i++)
	{
		glPushMatrix();
		glMultMatrixd(object_list[i]->transformation_matrix.m);
		glCallList(object_list[i]->list_name);
		glPopMatrix();
	}

	glGetDoublev(GL_PROJECTION_MATRIX, projection_matrix.m);
	glGetIntegerv(GL_VIEWPORT, viewport);

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

	glFlush();

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_CULL_FACE);

	unsigned short *temp_shadow_map;
	float n = 1.0 / 65536.0;

	temp_shadow_map = new unsigned short [shadow_map_size];

	glReadBuffer(GL_BACK);
	glReadPixels(0, 0, shadow_map_x, shadow_map_y, GL_DEPTH_COMPONENT, GL_SHORT, temp_shadow_map);

	smallest = 0.0;
	largest = 0.0;

	shadow_matrix = projection_matrix * modelview_matrix;

// find nearest and furthest points on the depth map so can optimise the near/far clipping planes
// the next time it is rendered

	for (int i = 0; i < shadow_map_size; i++)
	{
		shadow_map[i] = temp_shadow_map[i] * n;

		if (shadow_map[i] < smallest) smallest = shadow_map[i];
		if (shadow_map[i] > largest) largest = shadow_map[i];
	}

	delete [] temp_shadow_map;
}

typedef struct quad
{
	vector vertex1;
	vector vertex2;
	vector vertex3;
	vector vertex4;

	vector normal;

	struct quad *nextptr;
} QUAD;

/*****************************************************************************/
// generateShadowVolume() creates the shadow volume for the light
/*****************************************************************************/

void light::generateShadowVolume(object **object_list, int object_count)
{
	// NULL function
}

void pointlight::generateShadowVolume(object **object_list, int object_count)
{
	int i, j, k, edge_count, q(0);
	edge *edge_list = NULL;
	polygon *polygon_list = NULL;
	QUAD *quads = NULL, *cur_quad = NULL;
	vector p1, p2, p3;
	
	for (i = 0; i < object_count; i++)
	{
// Do a visibility test for all the polygons in the object. needed to determine outline
// edges of the object

		object_list[i]->visibilityTest(position);
		
		polygon_list = object_list[i]->polygons;
		edge_list = object_list[i]->edges;
		edge_count = object_list[i]->edge_count;
		
		for (j = 0; j < edge_count; j++)
		{
			int makeQuad = FALSE;

// To determine if an edge is part of the shapes outline, the polygons that share the edge must
// be flagged differently for visibility

			if (edge_list[j].right_polygon != -1)
			{
				if (polygon_list[edge_list[j].left_polygon].visible != polygon_list[edge_list[j].right_polygon].visible)
					makeQuad = TRUE;
			} else makeQuad = TRUE;

			if (makeQuad == TRUE)
			{
				if (quads == NULL)
				{
					quads = new QUAD;
					cur_quad = quads;
					q++;
				} else {
					cur_quad->nextptr = new QUAD;
					cur_quad = cur_quad->nextptr;
					q++;
				}
				
				vector ray1, ray2, average_normal, normal;

// This is a simple, but effective way of determining whether an quad is facing in the 
// right direction or not, since front and back face information is vital in this method
// of generating shadows
// The vectors between the centre of the polygons, and the center of the edge are
// added together and compared to the normal of the new quad. If they are pointing in the
// same direction then the quad is ok, otherwise it needs to be flipped (direction changed)

				if (edge_list[j].right_polygon != -1)
				{
					p1 = (polygon_list[edge_list[j].left_polygon].vertex1 + polygon_list[edge_list[j].left_polygon].vertex2 + polygon_list[edge_list[j].left_polygon].vertex3) / 3.0;
					p2 = (polygon_list[edge_list[j].right_polygon].vertex1 + polygon_list[edge_list[j].right_polygon].vertex2 + polygon_list[edge_list[j].right_polygon].vertex3) / 3.0;
					p3 = (edge_list[j].vertex1 + edge_list[j].vertex2) / 2.0;

					average_normal = (p3 - p1) + (p3 - p2);
				} else {
					p1 = (polygon_list[edge_list[j].left_polygon].vertex1 + polygon_list[edge_list[j].left_polygon].vertex2 + polygon_list[edge_list[j].left_polygon].vertex3) / 3.0;
					p2 = (edge_list[j].vertex1 + edge_list[j].vertex2) / 2.0;

					average_normal = p2 - p1;
				}
				
				ray1 = edge_list[j].vertex1 - position;
				ray1.normalise();
				ray1 = ray1 * 1000.0;
				
				ray2 = edge_list[j].vertex2 - position;
				ray2.normalise();
				ray2 = ray2 * 1000.0;
				
				normal = (edge_list[j].vertex2 - edge_list[j].vertex1) % (ray1 - edge_list[j].vertex1);
				normal.normalise();
				
				if (dot(normal, average_normal) >= 0.0)
				{
					cur_quad->vertex1 = edge_list[j].vertex1;
					cur_quad->vertex2 = edge_list[j].vertex2;
					cur_quad->vertex3 = ray2;
					cur_quad->vertex4 = ray1;
					
					cur_quad->normal = normal;
					cur_quad->nextptr = NULL;
				} else {
					cur_quad->vertex4 = edge_list[j].vertex1;
					cur_quad->vertex3 = edge_list[j].vertex2;
					cur_quad->vertex2 = ray2;
					cur_quad->vertex1 = ray1;
					
					cur_quad->normal = normal;
					cur_quad->nextptr = NULL;
				}
			}
		}
	}

// the stencil buffer operations must be set up to increment the value in the stencil
// buffer everytime a front face is drawn, and then to decrement the value when a back
// face is drawn

	glEnable(GL_CULL_FACE);
				
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	glDepthMask(GL_FALSE);
				
	glStencilFunc(GL_ALWAYS, 0x1, 0xFFFFFFFF);
				
	glColor3f(0.0, 0.0, 0.0);
				
	glEnable(GL_STENCIL_TEST);

// First pass... draws the front faces

	glCullFace(GL_BACK);
	glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);

	glBegin(GL_QUADS);

	cur_quad = quads;

	while (cur_quad != NULL)
	{
		glVertex3f(cur_quad->vertex1.x, cur_quad->vertex1.y, cur_quad->vertex1.z);
		glVertex3f(cur_quad->vertex2.x, cur_quad->vertex2.y, cur_quad->vertex2.z);
		glVertex3f(cur_quad->vertex3.x, cur_quad->vertex3.y, cur_quad->vertex3.z);
		glVertex3f(cur_quad->vertex4.x, cur_quad->vertex4.y, cur_quad->vertex4.z);

		cur_quad = cur_quad->nextptr;
	}

	glEnd();

// Second pass... draws the back faces

	glCullFace(GL_FRONT);
	glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);

	glBegin(GL_QUADS);

	cur_quad = quads;

	while (cur_quad != NULL)
	{
		glVertex3f(cur_quad->vertex1.x, cur_quad->vertex1.y, cur_quad->vertex1.z);
		glVertex3f(cur_quad->vertex2.x, cur_quad->vertex2.y, cur_quad->vertex2.z);
		glVertex3f(cur_quad->vertex3.x, cur_quad->vertex3.y, cur_quad->vertex3.z);
		glVertex3f(cur_quad->vertex4.x, cur_quad->vertex4.y, cur_quad->vertex4.z);

		cur_quad = cur_quad->nextptr;
	}

	glEnd();

	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glDepthMask(GL_TRUE);

	QUAD *temp;

	cur_quad = quads;

	while (cur_quad != NULL)
	{
		temp = cur_quad->nextptr;
		delete cur_quad;
		cur_quad = temp;
	}
}

void spotlight::generateShadowVolume(object **object_list, int object_count)
{
	int i, j, k, edge_count, q(0);
	edge *edge_list = NULL;
	polygon *polygon_list = NULL;
	QUAD *quads = NULL, *cur_quad = NULL;
	vector p1, p2, p3;
	
	for (i = 0; i < object_count; i++)
	{
		object_list[i]->visibilityTest(position);
		
		polygon_list = object_list[i]->polygons;
		edge_list = object_list[i]->edges;
		edge_count = object_list[i]->edge_count;
		
		for (j = 0; j < edge_count; j++)
		{
			int makeQuad = FALSE;

			if (edge_list[j].right_polygon != -1)
			{
				if (polygon_list[edge_list[j].left_polygon].visible != polygon_list[edge_list[j].right_polygon].visible)
					makeQuad = TRUE;
			} else makeQuad = TRUE;

			if (makeQuad == TRUE)
			{
				if (quads == NULL)
				{
					quads = new QUAD;
					cur_quad = quads;
					q++;
				} else {
					cur_quad->nextptr = new QUAD;
					cur_quad = cur_quad->nextptr;
					q++;
				}
				
				vector ray1, ray2, average_normal, normal;
				
				if (edge_list[j].right_polygon != -1)
				{
					p1 = (polygon_list[edge_list[j].left_polygon].vertex1 + polygon_list[edge_list[j].left_polygon].vertex2 + polygon_list[edge_list[j].left_polygon].vertex3) / 3.0;
					p2 = (polygon_list[edge_list[j].right_polygon].vertex1 + polygon_list[edge_list[j].right_polygon].vertex2 + polygon_list[edge_list[j].right_polygon].vertex3) / 3.0;
					p3 = (edge_list[j].vertex1 + edge_list[j].vertex2) / 2.0;

					average_normal = (p3 - p1) + (p3 - p2);
				} else {
					p1 = (polygon_list[edge_list[j].left_polygon].vertex1 + polygon_list[edge_list[j].left_polygon].vertex2 + polygon_list[edge_list[j].left_polygon].vertex3) / 3.0;
					p2 = (edge_list[j].vertex1 + edge_list[j].vertex2) / 2.0;

					average_normal = p2 - p1;
				}
				
				ray1 = edge_list[j].vertex1 - position;
				ray1.normalise();
				ray1 = ray1 * 1000.0;
				
				ray2 = edge_list[j].vertex2 - position;
				ray2.normalise();
				ray2 = ray2 * 1000.0;
				
				normal = (edge_list[j].vertex2 - edge_list[j].vertex1) % (ray1 - edge_list[j].vertex1);
				normal.normalise();
				
				if (dot(normal, average_normal) >= 0.0)
				{
					cur_quad->vertex1 = edge_list[j].vertex1;
					cur_quad->vertex2 = edge_list[j].vertex2;
					cur_quad->vertex3 = ray2;
					cur_quad->vertex4 = ray1;
					
					cur_quad->normal = normal;
					cur_quad->nextptr = NULL;
				} else {
					cur_quad->vertex4 = edge_list[j].vertex1;
					cur_quad->vertex3 = edge_list[j].vertex2;
					cur_quad->vertex2 = ray2;
					cur_quad->vertex1 = ray1;
					
					cur_quad->normal = normal;
					cur_quad->nextptr = NULL;
				}
			}
		}
	}

	glEnable(GL_CULL_FACE);
				
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	glDepthMask(GL_FALSE);
				
	glStencilFunc(GL_ALWAYS, 0x1, 0xFFFFFFFF);
				
	glColor3f(0.0, 0.0, 0.0);
				
	glEnable(GL_STENCIL_TEST);
	glCullFace(GL_BACK);
	glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);

	glBegin(GL_QUADS);

	cur_quad = quads;

	while (cur_quad != NULL)
	{
		glVertex3f(cur_quad->vertex1.x, cur_quad->vertex1.y, cur_quad->vertex1.z);
		glVertex3f(cur_quad->vertex2.x, cur_quad->vertex2.y, cur_quad->vertex2.z);
		glVertex3f(cur_quad->vertex3.x, cur_quad->vertex3.y, cur_quad->vertex3.z);
		glVertex3f(cur_quad->vertex4.x, cur_quad->vertex4.y, cur_quad->vertex4.z);

		cur_quad = cur_quad->nextptr;
	}

	glEnd();

	glCullFace(GL_FRONT);
	glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);

	glBegin(GL_QUADS);

	cur_quad = quads;

	while (cur_quad != NULL)
	{
		glVertex3f(cur_quad->vertex1.x, cur_quad->vertex1.y, cur_quad->vertex1.z);
		glVertex3f(cur_quad->vertex2.x, cur_quad->vertex2.y, cur_quad->vertex2.z);
		glVertex3f(cur_quad->vertex3.x, cur_quad->vertex3.y, cur_quad->vertex3.z);
		glVertex3f(cur_quad->vertex4.x, cur_quad->vertex4.y, cur_quad->vertex4.z);

		cur_quad = cur_quad->nextptr;
	}

	glEnd();

	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glDepthMask(GL_TRUE);

	QUAD *temp;

	cur_quad = quads;

	while (cur_quad != NULL)
	{
		temp = cur_quad->nextptr;
		delete cur_quad;
		cur_quad = temp;
	}
}


void directedlight::generateShadowVolume(object **object_list, int object_count)
{
	// NULL function
}
