#include <windows.h>
#include <iostream.h>
#include <glut.h>
#include "3dmath.h"

bitmap *new_bitmap;
colour c(255, 0, 0, 255), a(100, 100, 100, 255);
pointlight main_light;
vector light_pos, aim;
double anglex = 0.0, angley = 0.0, depth = 2.0;
double sin_angle(0.0);

float WINDOW_WIDTH = 720.0, WINDOW_HEIGHT = 576.0;
int MOUSE_X_LAST = -1, MOUSE_Y_LAST = -1, MOUSE_MODE = 0;

int DRAW_LIGHTS = FALSE;
int DRAW_SHADOWS = TRUE;
int RENDER_MODE = 1;
int TURNTABLE = FALSE;

unsigned int image_plane = 0;

/*****************************************************************************/
// Finds the furthest vertex from the camera... useful for determining near and far clipping planes
/*****************************************************************************/

double findFurthestPoint(void)
{
	double furthest = 0.0;

	for (int i = 0; i < object_count; i++)
	{
		for (int j = 0; j < object_list[i]->vertex_count; j++)
		{
			double cur;
			vector v;

			v = object_list[i]->vertices[j] - _CAMERA;
			cur = v.mag();

			if (cur > furthest) furthest = cur;
		}
	}

	return furthest;
}

/*****************************************************************************/
// Initialises the scene (sets up lights, camera, OpenGL settings, etc
/*****************************************************************************/

void init(char *scene_file_name)
{
	colour amb(50, 50, 50, 255), dif(200, 200, 200, 255), spec(255, 255, 255, 255);
	colour white(200, 200, 200, 255);

	float ambientlight[4] = {0.5, 0.5, 0.5, 1.0};

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambientlight);

	_CAMERA.x = aim.x + (cos(anglex) * -cos(angley));
	_CAMERA.y = aim.y + (sin(angley));
	_CAMERA.z = aim.z + (sin(anglex) * -cos(angley));

	light_pos(0.0, 5.721, 0.0);

	main_light.setPosition(light_pos);
	main_light.setIntensity(white);
	_LIGHTS = &main_light;

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_NORMALIZE);
	glEnable(GL_AUTO_NORMAL);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	object_count = loadSceneFile(scene_file_name, &object_list);

	glClearColor(0.0, 0.0, 0.0, 1.0);
}

char namestring[100];

/*****************************************************************************/
// Draws an axis
/*****************************************************************************/

void drawAxis(float scale)
{
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	glBegin(GL_LINES);
	glColor3f(scale, 0.0, 0.0);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(scale, 0.0, 0.0);

	glColor3f(0.0, scale, 0.0);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(0.0, scale, 0.0);

	glColor3f(0.0, 0.0, scale);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(0.0, 0.0, scale);
	glEnd();
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
}

/*****************************************************************************/
// Saves the frame buffer to disk (ie, screen grab)
/*****************************************************************************/

void saveFrameBuffer(void)
{
	static int frame_num(1);
	bitmap image(WINDOW_WIDTH, WINDOW_HEIGHT);

	glReadPixels(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT, GL_RGB, GL_UNSIGNED_BYTE, image.bitmap_data);
	sprintf(namestring, "frame%04d.bmp", frame_num);
	WriteBMPFile(namestring, &image);
	delete [] image.bitmap_data;
	cout << "Outputing frame: " << frame_num << endl;
	frame_num++;
	if (frame_num % 360 == 0)
	{
		if (RENDER_MODE == 1) RENDER_MODE = 3;
		else RENDER_MODE = 1;
	}
}

/*****************************************************************************/
// Uses OpenGLs select mode to determine visible objects, non visible objects
// do not have their lighting textures re-rendered that frame
/*****************************************************************************/

void objectVisibilityTest(object **object_list, int object_count)
{
	unsigned int visible_buffer[1024];
	int buffer_size = 1024, hits;

	glSelectBuffer(buffer_size, visible_buffer);
	glRenderMode(GL_SELECT);

	glInitNames();
	glPushName(0);

	for (int i = 0; i < object_count; i++)
	{
		if (object_list[i]->visible == TRUE)
		{
			glLoadName(i + 1);
			glCallList(object_list[i]->list_name);
			object_list[i]->visible_in_camera = FALSE;
		}
	}

	glFlush();

	hits = glRenderMode(GL_RENDER);

	int ptr = 3;
	for (int i = 0; i < hits; i++)
	{
		object_list[visible_buffer[ptr] - 1]->visible_in_camera = TRUE;
		ptr += 4;
	}
}

/*****************************************************************************/
// Renders the scene in two passes, then combines the two images and displays
// First pass: Draws objects with lighting textures
// Second pass: Draws objects with colour textures
/*****************************************************************************/

void renderTwoPasses(object **object_list, int object_count)
{
	bitmap temp1(WINDOW_WIDTH, WINDOW_HEIGHT, 3), temp2(WINDOW_WIDTH, WINDOW_HEIGHT, 3),
		output(WINDOW_WIDTH, WINDOW_HEIGHT, 3);
	glDisable(GL_LIGHTING);

	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	gluPerspective(75.0, WINDOW_WIDTH / WINDOW_HEIGHT, 0.1, 100.0);
	glViewport(0.0, 0.0, WINDOW_WIDTH, WINDOW_HEIGHT);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	gluLookAt(_CAMERA.x , _CAMERA.y, _CAMERA.z, aim.x, aim.y, aim.z, 0.0, 1.0, 0.0);

	objectVisibilityTest(object_list, object_count);

	glColor3f(1.0, 1.0, 1.0);

	for (int i = 0; i < object_count; i++)
	{
		object_list[i]->displayShadedObject();
	}

	glFlush();

	glReadPixels(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT, GL_RGB, GL_UNSIGNED_BYTE, temp1.bitmap_data);

	for (int i = 0; i < object_count; i++)
	{
		object_list[i]->displayTexturedObject();
	}
	glFlush();

	glReadPixels(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT, GL_RGB, GL_UNSIGNED_BYTE, temp2.bitmap_data);

	int bmpsize = WINDOW_WIDTH * WINDOW_HEIGHT * 3;
	float n(1.0 / 255.0);

	for (int i = 0; i < bmpsize; i++) output.bitmap_data[i] = (temp1.bitmap_data[i] * n) * (temp2.bitmap_data[i] * n) * 255.0;

	glDrawPixels(WINDOW_WIDTH, WINDOW_HEIGHT, GL_RGB, GL_UNSIGNED_BYTE, output.bitmap_data);

	light *cur = _LIGHTS;

/*****************************************************************************/
// Generates the shadow volume for each light
/*****************************************************************************/

	while (cur != NULL)
	{
		if (DRAW_SHADOWS == TRUE)
		{
			cur->generateShadowVolume(object_list, object_count);
			
			glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glLoadIdentity();
			glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
			
			glPopMatrix();
			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();
			glLoadIdentity();
			
			glColor4f(0.0, 0.0, 0.0, 1.0);
			
			glStencilFunc(GL_NOTEQUAL, 0x0, 0xFFFFFFFFL);
			glStencilOp(GL_ZERO, GL_ZERO, GL_ZERO);
			
			glDisable(GL_DEPTH_TEST);
			glDisable(GL_CULL_FACE);
			glEnable(GL_STENCIL_TEST);
			
			glBegin(GL_QUADS);
			glVertex3f(-1.0, -1.0, -0.5);
			glVertex3f(1.0, -1.0, -0.5);
			glVertex3f(1.0, 1.0, -0.5);
			glVertex3f(-1.0, 1.0, -0.5);
			glEnd();
			
			glEnable(GL_DEPTH_TEST);
			glDisable(GL_STENCIL_TEST);
			
			glPopMatrix();
		}

		if (DRAW_LIGHTS == TRUE)
		{
			cur->drawLight();
		}

		cur = cur->nextptr;
	}
	glStencilFunc(GL_ALWAYS, 0x1, 0xFFFFFFFFL);

	drawAxis(1.0);
	
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	delete [] temp1.bitmap_data;
	delete [] temp2.bitmap_data;
	delete [] output.bitmap_data;

	glutSwapBuffers();
}

/*****************************************************************************/
// Renders the entire scene Toon shaded in one pass
/*****************************************************************************/

void renderToonShaded(object **object_list, int object_count)
{
	pointlight *cur_light = (pointlight*)_LIGHTS;
	light *cur = NULL;
	float light_pos[4];

	glDisable(GL_TEXTURE_2D);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_LIGHTING);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	gluPerspective(75.0, WINDOW_WIDTH / WINDOW_HEIGHT, 0.1, 100.0);
	glViewport(0.0, 0.0, WINDOW_WIDTH, WINDOW_HEIGHT);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	gluLookAt(_CAMERA.x , _CAMERA.y, _CAMERA.z, aim.x, aim.y, aim.z, 0.0, 1.0, 0.0);

	objectVisibilityTest(object_list, object_count);

	int i = 0;
	while (cur_light != NULL)
	{
		glEnable(GL_LIGHT0 + i);

		light_pos[0] = cur_light->position.x;
		light_pos[1] = cur_light->position.y;
		light_pos[2] = cur_light->position.z;
		light_pos[3] = 1.0;

		glLightfv(GL_LIGHT0 + i, GL_POSITION, light_pos);

		cur_light = (pointlight*)cur_light->nextptr;
		i++;
	}

	for (int i = 0; i < object_count; i++)
	{
		object_list[i]->displayToonShadedObject();
	}

	cur = _LIGHTS;

	while (cur != NULL)
	{
		if (DRAW_SHADOWS == TRUE)
		{
			cur->generateShadowVolume(object_list, object_count);
			
			glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glLoadIdentity();
			glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
			
			glPopMatrix();
			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();
			glLoadIdentity();
			
			glColor4f(0.0, 0.0, 0.0, 0.5);
			
			glStencilFunc(GL_NOTEQUAL, 0x0, 0xFFFFFFFFL);
			glStencilOp(GL_ZERO, GL_ZERO, GL_ZERO);
			
			glDisable(GL_DEPTH_TEST);
			glDisable(GL_CULL_FACE);
			glEnable(GL_STENCIL_TEST);
			
			glBegin(GL_QUADS);
			glVertex3f(-1.0, -1.0, -0.5);
			glVertex3f(1.0, -1.0, -0.5);
			glVertex3f(1.0, 1.0, -0.5);
			glVertex3f(-1.0, 1.0, -0.5);
			glEnd();
			
			glEnable(GL_DEPTH_TEST);
			glDisable(GL_STENCIL_TEST);
			
			glPopMatrix();
		}

		if (DRAW_LIGHTS == TRUE)
		{
			cur->drawLight();
		}

		cur = cur->nextptr;
	}

	glStencilFunc(GL_ALWAYS, 0x1, 0xFFFFFFFFL);

	drawAxis(1.0);
	
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glutSwapBuffers();
}

/*****************************************************************************/
// Renders the scene using normal OpenGL lighting
/*****************************************************************************/

void renderOpenGL(object **object_list, int object_count)
{
	pointlight *cur_light = (pointlight*)_LIGHTS;
	light *cur = NULL;
	float light_pos[4];

	glDisable(GL_TEXTURE_2D);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_LIGHTING);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	gluPerspective(75.0, WINDOW_WIDTH / WINDOW_HEIGHT, 0.1, 100.0);
	glViewport(0.0, 0.0, WINDOW_WIDTH, WINDOW_HEIGHT);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	gluLookAt(_CAMERA.x , _CAMERA.y, _CAMERA.z, aim.x, aim.y, aim.z, 0.0, 1.0, 0.0);

	objectVisibilityTest(object_list, object_count);

	int i = 0;
	while (cur_light != NULL)
	{
		glEnable(GL_LIGHT0 + i);

		light_pos[0] = cur_light->position.x;
		light_pos[1] = cur_light->position.y;
		light_pos[2] = cur_light->position.z;
		light_pos[3] = 1.0;

		glLightfv(GL_LIGHT0 + i, GL_POSITION, light_pos);

		cur_light = (pointlight*)cur_light->nextptr;
		i++;
	}

	for (int i = 0; i < object_count; i++)
	{
		object_list[i]->displayOpenGLShadedObject();
	}

	cur = _LIGHTS;

	glDisable(GL_LIGHTING);

	while (cur != NULL)
	{
		if (DRAW_SHADOWS == TRUE)
		{
			cur->generateShadowVolume(object_list, object_count);
			
			glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glLoadIdentity();
			glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
			
			glPopMatrix();
			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();
			glLoadIdentity();
			
			glColor4f(0.0, 0.0, 0.0, 0.5);
			
			glStencilFunc(GL_NOTEQUAL, 0x0, 0xFFFFFFFFL);
			glStencilOp(GL_ZERO, GL_ZERO, GL_ZERO);
			
			glDisable(GL_DEPTH_TEST);
			glDisable(GL_CULL_FACE);
			glEnable(GL_STENCIL_TEST);
			
			glBegin(GL_QUADS);
			glVertex3f(-1.0, -1.0, -0.5);
			glVertex3f(1.0, -1.0, -0.5);
			glVertex3f(1.0, 1.0, -0.5);
			glVertex3f(-1.0, 1.0, -0.5);
			glEnd();
			
			glEnable(GL_DEPTH_TEST);
			glDisable(GL_STENCIL_TEST);
			
			glPopMatrix();
		}

		if (DRAW_LIGHTS == TRUE)
		{
			cur->drawLight();
		}

		cur = cur->nextptr;
	}

	glStencilFunc(GL_ALWAYS, 0x1, 0xFFFFFFFFL);

	drawAxis(1.0);
	
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glutSwapBuffers();
}

#define BLUR 10.0

/*****************************************************************************/
// Main display function... determines which mode to render in, and re-positions camera
/*****************************************************************************/

void display(void)
{
	if (TURNTABLE == TRUE)
	{
		anglex += PI / 180.0;
		angley = 0.5;
	}

	_CAMERA.x = aim.x + (cos(anglex) * -cos(angley)) * 10.0;
	_CAMERA.y = aim.y + (sin(angley)) * 10.0;
	_CAMERA.z = aim.z + (sin(anglex) * -cos(angley)) * 10.0;

	if (RENDER_MODE == 1) renderOpenGL(object_list, object_count);
	else if (RENDER_MODE == 2) renderTwoPasses(object_list, object_count);
	else renderToonShaded(object_list, object_count);
}

/*****************************************************************************/
// Updates the display when idle (recalls the display function
/*****************************************************************************/

void update(void)
{
	glutPostRedisplay();
}

/*****************************************************************************/
// Stores the new window size in global variables that are used to determine viewport
// and perspective settings
/*****************************************************************************/

void reshape(int x, int y)
{
	WINDOW_WIDTH = x;
	WINDOW_HEIGHT = y;

	glutPostRedisplay();
}

/*****************************************************************************/
// Moves the camera whenever the mouse button is pressed and dragged in the window
/*****************************************************************************/

void motion(int x, int y)
{
	vector move = aim - _CAMERA;
	if (MOUSE_X_LAST == -1)
	{
		MOUSE_X_LAST = x;
		MOUSE_Y_LAST = y;
	}

	switch (MOUSE_MODE)
	{
	case 1:
		anglex += ((x - MOUSE_X_LAST) / WINDOW_WIDTH) * 2;
		angley += ((y - MOUSE_Y_LAST) / WINDOW_HEIGHT) * 2;
		break;
	case 2:
		break;
	case 3:
		aim = aim + (move * -((y - MOUSE_Y_LAST) / WINDOW_HEIGHT) * 2);
		_CAMERA = _CAMERA + (move * -((y - MOUSE_Y_LAST) / WINDOW_HEIGHT) * 2);
		break;
	default:
		break;
	}

	MOUSE_X_LAST = x;
	MOUSE_Y_LAST = y;

	glutPostRedisplay();
}

/*****************************************************************************/
// Needed to help with motion() function
/*****************************************************************************/

void mouse(int button, int state, int x, int y)
{
	if (state == GLUT_UP)
	{
		MOUSE_MODE = 0;
	} else {
		if (button == GLUT_LEFT_BUTTON)
		{
			MOUSE_MODE = 1;
			MOUSE_X_LAST = -1;
			MOUSE_Y_LAST = -1;
		} else if (button == GLUT_MIDDLE_BUTTON)
		{
			MOUSE_MODE = 2;
			MOUSE_X_LAST = -1;
			MOUSE_Y_LAST = -1;
		} else if (button == GLUT_RIGHT_BUTTON)
		{
			MOUSE_MODE = 3;
			MOUSE_X_LAST = -1;
			MOUSE_Y_LAST = -1;
		}
	}
}

/*****************************************************************************/
// Determines which keys have been pressed and which mode to go into
/*****************************************************************************/

void keyboard(unsigned char key, int x, int y)
{
	switch(key)
	{
	case 's': case 'S':
		if (CALC_SPECULAR_MAP_REALTIME == TRUE) CALC_SPECULAR_MAP_REALTIME = FALSE;
		else CALC_SPECULAR_MAP_REALTIME = TRUE;
		break;
	case 'a': case 'A':
		if (CALC_AMBIENT_MAP_REALTIME == TRUE) CALC_AMBIENT_MAP_REALTIME = FALSE;
		else CALC_AMBIENT_MAP_REALTIME = TRUE;
		break;
	case 'd': case 'D':
		if (CALC_DIFFUSE_MAP_REALTIME == TRUE) CALC_DIFFUSE_MAP_REALTIME = FALSE;
		else CALC_DIFFUSE_MAP_REALTIME = TRUE;
		break;
	case 'l': case 'L':
		if (DRAW_LIGHTS == TRUE) DRAW_LIGHTS = FALSE;
		else DRAW_LIGHTS = TRUE;
		break;
	case 'h': case 'H':
		if (DRAW_SHADOWS == TRUE) DRAW_SHADOWS = FALSE;
		else DRAW_SHADOWS = TRUE;
		break;
	case 'f': case 'F':
		char sname[1000];
		for (int i = 0; i < object_count; i++) {
			sprintf(sname, "diffuse%02d.bmp", i);
			object_list[i]->saveTextureBuffer(DIFFUSE, sname);
			sprintf(sname, "ambient%02d.bmp", i);
			object_list[i]->saveTextureBuffer(AMBIENT, sname);
			sprintf(sname, "specular%02d.bmp", i);
			object_list[i]->saveTextureBuffer(SPECULAR, sname);
			sprintf(sname, "alpha%02d.bmp", i);
			object_list[i]->saveTextureBuffer(ALPHA, sname);
			sprintf(sname, "normals%02d.bmp", i);
			object_list[i]->saveTextureBuffer(NORMALS, sname);
			sprintf(sname, "positions%02d.bmp", i);
			object_list[i]->saveTextureBuffer(POSITIONS, sname); }
		break;
	case 'c': case 'C':
		saveFrameBuffer();
		break;
	case 't': case 'T':
		if (RENDER_MODE == 1) RENDER_MODE = 2;
		else if (RENDER_MODE == 2) RENDER_MODE = 3;
		else RENDER_MODE = 1;
		break;
	default:
		break;
	}
}

main(int argc, char **argv)
{
	glutInit(&argc, argv);

	printf("%s\n", gluGetString(GLU_VERSION));
	printf("%s\n", gluGetString(GLU_EXTENSIONS));
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STENCIL);
	glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);
	glutInitWindowPosition(100, 100);
	glutCreateWindow(argv[1]);

	if (argc >= 2)
	{
		init(argv[1]);
		if (argc == 3)
		{
			if (strcmp("-turn", argv[2]) == 0) TURNTABLE = TRUE;
			cout << "*************************************" << endl;
			cout << "*************************************" << endl;
			cout << "*************************************" << endl << endl;
			cout << "Entering Turntable Mode" << endl;
			cout << "*************************************" << endl << endl;
			cout << "*************************************" << endl;
			cout << "*************************************" << endl;
		}
	} else {
		init("test2.txt");
	}

	glutDisplayFunc(display);
	glutIdleFunc(update);
	glutReshapeFunc(reshape);

	if (TURNTABLE == FALSE)
	{
		glutMotionFunc(motion);
		glutMouseFunc(mouse);
	}

	glutKeyboardFunc(keyboard);

	glutMainLoop();

	return 0;
}